/*
Copyright (C) 2007 JJ Foote

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <nds.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include "MessageQueue.h"
#include "802.11.h"
#include "lobby.h"
#include "dice.h"

#define YAHTZEE_GAMECODE 6183
#define MAX_CLIENTS 6
#define MAX_DATA 128

#define MESSAGE_GAME_JOIN 1
#define MESSAGE_GAME_START 2
#define MESSAGE_TURN_START 3
#define MESSAGE_ROLL_DICE 4
#define MESSAGE_SCORE_TURN 5
#define MESSAGE_WAITING_FOR 6
#define MESSAGE_NEED_DICE 7
#define MESSAGE_END_GAME 8

#define MODE_MENU 0
#define MODE_WAITING 1
#define MODE_PLAYING 2
#define MODE_WAITING_FOR_DICE 3
#define MODE_NEED_CLIENTS 4

#define SCORE_ONES 0
#define SCORE_TWOS 1
#define SCORE_THREES 2
#define SCORE_FOURS 3
#define SCORE_FIVES 4
#define SCORE_SIXES 5
#define SCORE_THREE_OF_A_KIND 6
#define SCORE_FOUR_OF_A_KIND 7
#define SCORE_FULL_HOUSE 8
#define SCORE_SMALL_STRAIGHT 9
#define SCORE_LARGE_STRAIGHT 10
#define SCORE_YAHTZEE 11
#define SCORE_CHANCE 12

#define SOURCE_FIRMWARE 0
#define SOURCE_LOBBY 1
#define SOURCE_GENERIC 2

u16 bg0map[32 * 24];
u8 bg3map[256 * 192];

unsigned char send_buffer[MAX_DATA];
int i_am_server = 0;
LPLOBBY_USER server;
LPLOBBY_USER clients[MAX_CLIENTS];
LPLOBBY_ROOM game_room;

typedef struct player_struct {
	char name[32];
	LPLOBBY_USER client;
	int ones, twos, threes, fours, fives, sixes;
	int three_of_a_kind, four_of_a_kind, full_house;
	int small_straight, large_straight, yahtzee, chance;
	struct player_struct *next, *prev;
} player_t;

int wifi_active = 0, lobby_ever_init = 0;
int mode;
int dice[5];
int held[5];
int roll;
time_t dice_animation_start;
int current_client, current_player;
int num_players, num_clients, num_local_players;
int players_per_client[MAX_CLIENTS];
player_t *player, *now;

void roll_dice();
void advance_player();
void start_game(int is_local);
player_t *init_players(int source);
void end_game();
void join_game(LPLOBBY_ROOM room);
void host_game();
void draw_scoresheet();
void draw_server_screen();
void draw_menu();
void handle_input();
void handle_menu_input();
void handle_server_input();
void score_turn(int category);
void display_final_scores(int *scores);
void draw_bitmap(int x, int y, int img);
void activate_wifi();
int animate_dice(time_t start);

// rolling / scoring macros
#define DICE_COUNT(x) ((dice[0] == (x) ? 1 : 0) + (dice[1] == (x) ? 1 : 0) + (dice[2] == (x) ? 1 : 0) \
	+ (dice[3] == (x) ? 1 : 0) + (dice[4] == (x) ? 1 : 0))

#define SUM_DICE(x) ((dice[0] == (x) ? (x) : 0) + (dice[1] == (x) ? (x) : 0) + (dice[2] == (x) ? (x) : 0) \
	+ (dice[3] == (x) ? (x) : 0) + (dice[4] == (x) ? (x) : 0))

#define SUM_ALL_DICE (dice[0] + dice[1] + dice[2] + dice[3] + dice[4])

#define DICE_NUM_PAIRS ((DICE_COUNT(1) >= 2 ? 1 : 0) + (DICE_COUNT(2) >= 2 ? 1 : 0) + (DICE_COUNT(3) >= 2 ? 1 : 0) \
	+ (DICE_COUNT(4) >= 2 ? 1 : 0) + (DICE_COUNT(5) >= 2 ? 1 : 0) + (DICE_COUNT(6) >= 2 ? 1 : 0))

#define DICE_IS_YAHTZEE (dice[0] == dice[1] && dice[1] && dice[2] && dice[2] == dice[3] && dice[3] == dice[4])

#define DICE_IS_THREE_OF_A_KIND (DICE_COUNT(1) >= 3 || DICE_COUNT(2) >= 3 || DICE_COUNT(3) >= 3 || DICE_COUNT(4) >= 3 \
	|| DICE_COUNT(5) >= 3 || DICE_COUNT(6) >= 3 ? 1 : 0)

#define DICE_IS_FOUR_OF_A_KIND (DICE_COUNT(1) >= 4 || DICE_COUNT(2) >= 4 || DICE_COUNT(3) >= 4 || DICE_COUNT(4) >= 4 \
	|| DICE_COUNT(5) >= 4 || DICE_COUNT(6) >= 4 ? 1 : 0)

#define DICE_IS_FULL_HOUSE (DICE_IS_THREE_OF_A_KIND && DICE_NUM_PAIRS >= 2 ? 1 : 0)

#define DICE_IS_SMALL_STRAIGHT (DICE_COUNT(3) && DICE_COUNT(4) && ((DICE_COUNT(5) && DICE_COUNT(6)) || (DICE_COUNT(2) && DICE_COUNT(5)) \
	|| (DICE_COUNT(1) && DICE_COUNT(2))) ? 1 : 0)

#define DICE_IS_LARGE_STRAIGHT (DICE_COUNT(2) && DICE_COUNT(3) && DICE_COUNT(4) && DICE_COUNT(5) && (DICE_COUNT(1) || DICE_COUNT(6)) ? 1 : 0)

#define TOP_SUBTOTAL ((now->ones != -1 ? now->ones : 0) + (now->twos != -1 ? now->twos : 0) + (now->threes != -1 ? now->threes : 0) \
	+ (now->fours != -1 ? now->fours : 0) + (now->fives != -1 ? now->fives : 0) + (now->sixes != -1 ? now->sixes : 0))

#define BOTTOM_SUBTOTAL ((now->three_of_a_kind != -1 ? now->three_of_a_kind : 0) + (now->four_of_a_kind != -1 ? now->four_of_a_kind : 0) \
	+ (now->full_house != -1 ? now->full_house : 0) + (now->small_straight != -1 ? now->small_straight : 0) \
	+ (now->large_straight != -1 ? now->large_straight : 0) + (now->yahtzee != -1 ? now->yahtzee : 0) + (now->chance != -1 ? now->chance : 0))

//---------------------------------------------------------------------------------
#define allowed_to_sleep (*(volatile int*)0x027FDF00)
void enter_sleep_mode()
{
	if(!allowed_to_sleep)
		return;

	/* save the current IRQs, and turn off everything except VBlank */
	unsigned long oldIE = REG_IE;
	REG_IE = IRQ_VBLANK;

	/* power down the display */
	powerOFF(POWER_LCD);

	/* wait until the VBLANK (triggered by the ARM7 waking up) wakes us up */
	swiIntrWait(1, IRQ_VBLANK); 

	/* wait a bit longer until returning power */
	while(REG_VCOUNT != 0);
	while(REG_VCOUNT == 0);
	while(REG_VCOUNT != 0);

	/* restore power, and restore IRQs */
	powerON(POWER_LCD);
	REG_IE = oldIE;
}

//---------------------------------------------------------------------------------
u32 wait_for_keypress(u32 keymask)
{
	u32 kd;

	// wait until they've let go
	while(keysHeld() & keymask) { scanKeys(); swiWaitForVBlank(); }

	// wait until they pressed it
	do {
	   	swiWaitForVBlank();
		scanKeys();
		kd = keysDown();

		if(keysHeld() & KEY_LID) {
			enter_sleep_mode();
		}
	} while(!(kd & keymask));

	// wait until they release it
	while(keysHeld() & kd) { scanKeys(); swiWaitForVBlank(); }

	// return the key(s) pressed
	return kd;
}

//---------------------------------------------------------------------------------
void flip()
{
	DC_FlushRange(bg0map, 32 * 24 * 2);
	dmaCopy(bg0map, (u16*)SCREEN_BASE_BLOCK(0), 32 * 24 * 2);
	DC_FlushRange(bg3map, 256 * 192);
	dmaCopy(bg3map, (u16*)BG_BMP_RAM(2), 256 * 192);
}

void main_screen_clear()
{
	int i;

	for(i = 0; i < 32 * 24; i++) {
		bg0map[i] = (15 << 12) | (u16)' ';
	}

	memset(bg3map, 0, 256 * 192);
}

void main_screen_print_char(int col, int row, int ch)
{
	bg0map[col + row * 32] = (15 << 12) | (u16)ch;
}

void main_screen_print(int col, int row, char *str)
{
	while(*str != '\0') {
		if(*str == '\n') {
			col = 0;
			row++;
		} else {
			main_screen_print_char(col, row, *str);
		}

		col++;
		if(col >= 32) {
			col = 0;
			row++;
		}

		str++;
	}
}

void main_screen_printf(int col, int row, char *fmt, ...)
{
	static char buf[1024];
	va_list args;

	va_start(args, fmt);
	vsnprintf(buf, 1024, fmt, args); buf[1023] = '\0';
	va_end(args);

	main_screen_print(col, row, buf);
}

void main_screen_colorize(int col, int row, int width, int palette)
{
	int i;
	for(i = 0; i < width; i++) {
		bg0map[row * 32 + col + i] &= (palette << 12) | (0xFFF);
	}
}

static inline void console_gotoxy(int col, int row)
{
	iprintf("\x1b[%d;%dH", row, col);
}

void sub_screen_printf(int col, int row, char *fmt, ...)
{
	static char buf[1024];
	va_list args;

	va_start(args, fmt);
	vsnprintf(buf, 1024, fmt, args); buf[1023] = '\0';
	va_end(args);

	console_gotoxy(col, row);
	iprintf(buf);
}

//---------------------------------------------------------------------------------
void VBlankIRQ(void)
{
	if(wifi_active) {
		REG_IF = IRQ_VBLANK;
		IPC_RcvCompleteCheck();
		LOBBY_Update();
	}
}

//---------------------------------------------------------------------------------
void receiveUserData(unsigned char *data, int length, LPLOBBY_USER from)
{
	switch(data[0]) {
		case MESSAGE_GAME_START:
			num_players = data[1];
			if(!i_am_server) {
				// the clients don't need to know the arrangement of other clients
				num_clients = 1;
				players_per_client[0] = INT_MAX;
				clients[0] = NULL;
				init_players(SOURCE_GENERIC); // we'll learn each user's real name on his/her first turn
				now = player;
				mode = MODE_WAITING;
			}
			break;

		case MESSAGE_GAME_JOIN:
			if(i_am_server) {
				players_per_client[data[1]] = data[2];
			} else {
				mode = MODE_WAITING;
			}
			break;

		case MESSAGE_TURN_START:
			//iprintf("It's my turn now\n");
			mode = MODE_PLAYING;
			roll = 0;
			held[0] = 0;
			held[1] = 0;
			held[2] = 0;
			held[3] = 0;
			held[4] = 0;
			break;

		case MESSAGE_WAITING_FOR:
			if(!i_am_server) {
				now = now->next;
				snprintf(now->name, 32, "%s", data + 1); now->name[31] = '\0';
				roll = 0;
				mode = MODE_WAITING;
				//iprintf("Waiting for %s\n", now->name);
			}
			break;

		case MESSAGE_NEED_DICE:
			held[0] = (data[1] ? 1 : 0);
			held[1] = (data[2] ? 1 : 0);
			held[2] = (data[3] ? 1 : 0);
			held[3] = (data[4] ? 1 : 0);
			held[4] = (data[5] ? 1 : 0);
			roll_dice();
			break;

		case MESSAGE_ROLL_DICE:
			if(roll >= 3) {
				//iprintf("Cheater! Only 3 rolls!\n"); while(1);
			} else if(!i_am_server) {
				dice[0] = data[1];
				dice[1] = data[2];
				dice[2] = data[3];
				dice[3] = data[4];
				dice[4] = data[5];

				roll++;

				dice_animation_start = time(0) + 1;

				if(mode == MODE_WAITING_FOR_DICE) mode = MODE_PLAYING;
			}
			break;

		case MESSAGE_SCORE_TURN:
			//iprintf("score me\n");
			score_turn(data[1]);
			if(i_am_server)
				advance_player();
			break;

		case MESSAGE_END_GAME:
			if(!i_am_server)
				display_final_scores((int*) &data[4]);
			break;
	}
}

void userCallback(LPLOBBY_USER user, unsigned long reason)
{
#if 0
	switch(reason)
	{
		case USERINFO_REASON_REGOGNIZE:
			iprintf("user recognized %s    \n", LOBBY_GetUserName(user));
			break;
		case USERINFO_REASON_TIMEOUT:
			iprintf("user timeout %s       \n", LOBBY_GetUserName(user));
			break;
		case USERINFO_REASON_RETURN:
			iprintf("user returned %s      \n", LOBBY_GetUserName(user));
			break;
		case USERINFO_REASON_ROOMCREATED:
			iprintf("user created room %s  \n", LOBBY_GetUserName(user));
			break;
		case USERINFO_REASON_ROOMCHANGE:
			iprintf("user changed room %s  \n", LOBBY_GetUserName(user));
			break;
	}
#endif
}

//---------------------------------------------------------------------------------
void host_game()
{
	char temp[32], firmware_name[32];
	int i;

	for(i = 0; i < PersonalData->nameLen; i++)
		firmware_name[i] = PersonalData->name[i] & 0xFF; // cheap UTF-16 conversion
	firmware_name[i] = '\0';

	snprintf(temp, 32, "YahtzeeDS - %s", firmware_name); temp[31] = '\0';
	i_am_server = 1;

	if(wifi_active) {
		LOBBY_CreateRoom(temp, MAX_CLIENTS, YAHTZEE_GAMECODE, (int)(atof(VERSION) * 100));
		game_room = LOBBY_GetRoomByID(ROOMID_MYROOM);
		server = LOBBY_GetUserByID(USERID_MYSELF);
	}

	players_per_client[0] = num_local_players;
}

void join_game(LPLOBBY_ROOM room)
{
	if(!wifi_active)
		return;

	game_room = room;

	LOBBY_JoinRoom(game_room);
	i_am_server = 0;
	server = LOBBY_GetRoomUserBySlot(game_room, 0);

	// stupid hack to find my slot in the room
	LPLOBBY_USER me = LOBBY_GetUserByID(USERID_MYSELF);
	int i, max;
	do {
		swiWaitForVBlank();

		max = LOBBY_GetUsercountInRoom(game_room);

		for(i = 0; i < max; i++) {
		   if(LOBBY_GetRoomUserBySlot(game_room, i) == me) {
			   break;
		   }
		}
	} while(i == max);

	// tell the server we're here
	memset(send_buffer, 0, MAX_DATA);
	send_buffer[0] = MESSAGE_GAME_JOIN;
	send_buffer[1] = i;
	send_buffer[2] = num_local_players;
	LOBBY_SendToUser(server, 0x0001, send_buffer, MAX_DATA);

	main_screen_clear();
	main_screen_printf(0,0, "Waiting for the game to start.");
	flip();
}

player_t *init_players(int source)
{
	player_t *temp = NULL, *prev = NULL;
	char firmware_name[32];
	char *name;
	int i, j;

	prev = NULL;
	for(i = 0, current_player = 0, current_client = 0; i < num_players; i++) {
		temp = malloc(sizeof(player_t));
		if(temp == NULL) { iprintf("Out of memory!\n"); while(1); }

		if(source == SOURCE_FIRMWARE) {
			for(j = 0; j < PersonalData->nameLen; j++)
				firmware_name[j] = PersonalData->name[j] & 0xFF; // cheap UTF-16 conversion
			firmware_name[j] = '\0';
			name = firmware_name;
		} else if(source == SOURCE_LOBBY) {
			name = (char*) LOBBY_GetUserName(clients[current_client]);
		} else {
			snprintf(firmware_name, 32, "%s", "Player"); firmware_name[31] = '\0';
			name = firmware_name;
		}

		if(current_player == 0) {
			snprintf(temp->name, 32, "%s", name); temp->name[31] = '\0';
		} else {
			snprintf(temp->name, 28, "%s", name); temp->name[27] = '\0';
			snprintf(temp->name, 32, "%s #%d", temp->name, current_player + 1); temp->name[31] = '\0';
		}

		temp->client = clients[current_client];
		temp->ones = -1;
		temp->twos = -1;
		temp->threes = -1;
		temp->fours = -1;
		temp->fives = -1;
		temp->sixes = -1;
		temp->three_of_a_kind = -1;
		temp->four_of_a_kind = -1;
		temp->full_house = -1;
		temp->small_straight = -1;
		temp->large_straight = -1;
		temp->yahtzee = -1;
		temp->chance = -1;

		if(prev == NULL)
			player = temp;
		else
			prev->next = temp;

		temp->prev = prev;

		prev = temp;

		current_player++;
		if(current_player >= players_per_client[current_client]) {
			 current_player = 0;
			 current_client++;
		}
	}

	prev->next = player;
	player->prev = prev;

	return prev;
}

void start_game(int is_local)
{
	player_t *prev;
	int i;

	if(is_local) {
		num_clients = 1;
		num_players = num_local_players;
		players_per_client[0] = INT_MAX;
		server = NULL;
		i_am_server = 1;
		clients[0] = NULL;

		// for a single player game, use the DS's firmware name; otherwise, use generic Player 1, Player 2, etc.
		prev = init_players((num_players == 1 ? SOURCE_FIRMWARE : SOURCE_GENERIC));
	} else {
		num_players = 0;
		num_clients = 0;
		for(i = 0; i < LOBBY_GetUsercountInRoom(game_room) && i < MAX_CLIENTS; i++) {
			if(players_per_client[i] > 0) {
				clients[num_clients] = LOBBY_GetRoomUserBySlot(game_room, i);
				num_clients++;
				num_players += players_per_client[i];
			}
		}

		prev = init_players(SOURCE_LOBBY);

		memset(send_buffer, 0, MAX_DATA);
		send_buffer[0] = MESSAGE_GAME_START;
		send_buffer[1] = num_players;
		LOBBY_SendToRoom(game_room, 0x0001, send_buffer, MAX_DATA);
	}

	current_client = 0;
	current_player = 0;
	now = prev;

	/*
	for(i = 0; i < num_clients; i++)
		iprintf("%d) 0x%x - %d players\n", i, clients[i], players_per_client[i]);
	for(i = 0, prev = player; i < num_players; i++, prev = prev->next)
		iprintf("%d) 0x%x %s\n", i, prev->client, prev->name);
	*/

	now = NULL;
	advance_player();
}

void end_game()
{
	int i;
	player_t *temp, *dead;

	dead = player;
	for(i = 0; i < num_players; i++) {
		temp = dead->next;
		free(dead);
		dead = temp;
	}
}

void advance_player()
{
	int *ptr;

	if(now == NULL) {
		now = player;
	} else {
		if(now->client == now->next->client) {
			current_player++;
		} else {
			current_client++;
			current_player = 0;
		}
		now = now->next;
	}

	if(now->ones != -1 && now->twos != -1 && now->threes != -1 && now->fours != -1 && now->fives != -1 && now->sixes != -1
		&& now->three_of_a_kind != -1 && now->four_of_a_kind != -1 && now->full_house != -1 && now->small_straight != -1
		&& now->large_straight != -1 && now->yahtzee != -1 && now->chance != -1)
	{
		memset(send_buffer, 0, MAX_DATA);
		send_buffer[0] = MESSAGE_END_GAME;
		ptr = (int*) &send_buffer[4];
		do {
			*ptr = TOP_SUBTOTAL + (TOP_SUBTOTAL >= 63 ? 35 : 0) + BOTTOM_SUBTOTAL;
			now = now->next;
		} while(now != player);

		if(num_clients > 1) {
			LOBBY_SendToRoom(game_room, 0x0001, send_buffer, MAX_DATA);
			display_final_scores((int*) &send_buffer[4]);
		} else {
			display_final_scores((int*) &send_buffer[4]);
			return;
		}
	}

	if(now->client == server) {
		//iprintf("It's my turn now\n");

		// local player
		mode = MODE_PLAYING;
		roll = 0;
		held[0] = 0;
		held[1] = 0;
		held[2] = 0;
		held[3] = 0;
		held[4] = 0;

		if(num_clients > 1) {
			memset(send_buffer, 0, MAX_DATA);
			send_buffer[0] = MESSAGE_WAITING_FOR;
			sprintf((char*)send_buffer + 1, "%s", now->name);
			LOBBY_SendToRoom(game_room, 0x0001, send_buffer, MAX_DATA);
		}
	} else {
		//iprintf("Waiting for %s\n", now->name);

		roll = 0;

		// remote player, tell them to go
		memset(send_buffer, 0, MAX_DATA);
		send_buffer[0] = MESSAGE_WAITING_FOR;
		sprintf((char*)send_buffer + 1, "%s", now->name);
		LOBBY_SendToRoom(game_room, 0x0001, send_buffer, MAX_DATA);

		memset(send_buffer, 0, MAX_DATA);
		send_buffer[0] = MESSAGE_TURN_START;
		send_buffer[1] = current_player;
		LOBBY_SendToUser(now->client, 0x0001, send_buffer, MAX_DATA);
		mode = MODE_WAITING;
	}
}

void roll_dice()
{
	int i;

	if(i_am_server) {
		dice_animation_start = time(0);

		for(i = 0; i < 5; i++) {
			if(!held[i])
				dice[i] = rand() % 6 + 1;
		}

		roll++;

		if(num_clients > 1) {
			memset(send_buffer, 0, MAX_DATA);
			send_buffer[0] = MESSAGE_ROLL_DICE;
			for(i = 0; i < 5; i++) {
				send_buffer[i + 1] = dice[i];
			}
			LOBBY_SendToRoom(game_room, 0x0001, send_buffer, MAX_DATA);
		}
	} else {
		memset(send_buffer, 0, MAX_DATA);
		send_buffer[0] = MESSAGE_NEED_DICE;
		send_buffer[1] = held[0];
		send_buffer[2] = held[1];
		send_buffer[3] = held[2];
		send_buffer[4] = held[3];
		send_buffer[5] = held[4];
		LOBBY_SendToUser(server, 0x0001, send_buffer, MAX_DATA);
		mode = MODE_WAITING_FOR_DICE;

		dice_animation_start = time(0);
	}
}

void score_turn(int category)
{
	int score = 0;
	int *loc = NULL;

	switch(category) {
		case SCORE_ONES:
			loc = &now->ones;
			score = SUM_DICE(1);
			break;
		case SCORE_TWOS:
			loc = &now->twos;
			score = SUM_DICE(2);
			break;
		case SCORE_THREES:
			loc = &now->threes;
			score = SUM_DICE(3);
			break;
		case SCORE_FOURS:
			loc = &now->fours;
			score = SUM_DICE(4);
			break;
		case SCORE_FIVES:
			loc = &now->fives;
			score = SUM_DICE(5);
			break;
		case SCORE_SIXES:
			loc = &now->sixes;
			score = SUM_DICE(6);
			break;
		case SCORE_THREE_OF_A_KIND:
			loc = &now->three_of_a_kind;
			score = (DICE_IS_THREE_OF_A_KIND ? SUM_ALL_DICE : 0);
			break;
		case SCORE_FOUR_OF_A_KIND:
			loc = &now->four_of_a_kind;
			score = (DICE_IS_FOUR_OF_A_KIND ? SUM_ALL_DICE : 0);
			break;
		case SCORE_FULL_HOUSE:
			loc = &now->full_house;
			score = (DICE_IS_FULL_HOUSE ? 25 : 0);
			break;
		case SCORE_SMALL_STRAIGHT:
			loc = &now->small_straight;
			score = (DICE_IS_SMALL_STRAIGHT ? 30 : 0);
			break;
		case SCORE_LARGE_STRAIGHT:
			loc = &now->large_straight;
			score = (DICE_IS_LARGE_STRAIGHT ? 40 : 0);
			break;
		case SCORE_YAHTZEE:
			loc = &now->yahtzee;
			score = (DICE_IS_YAHTZEE ? 50 : 0);
			break;
		case SCORE_CHANCE:
			loc = &now->chance;
			score = SUM_ALL_DICE;
			break;
		default:
			loc = NULL;
	}

	if(loc == NULL) {
		//iprintf("Cheater! You must pick a score.\n"); while(1);
	} else if(*loc != -1) {
		//iprintf("Cheater! You already scored that\n"); while(1);
	} else {
		if(DICE_IS_YAHTZEE && now->yahtzee >= 50)
			now->yahtzee += 100;
		*loc = score;
	}
}

void draw_scoresheet()
{
	int i;

	// show the previous player's score (assuming more than 1 player) on the sub-screen
	if(now->prev != now) {
		now = now->prev;

		for(i = 14; i < 24; i++)
			sub_screen_printf(0, i, "                                ");

		sub_screen_printf(0, 14, "%s", now->name);

		if(now->ones == -1)
			sub_screen_printf(0, 15, "ONES  :    ");
		else
			sub_screen_printf(0, 15, "ONES  :  %2d", now->ones);

		if(now->twos == -1)
			sub_screen_printf(0, 16, "TWOS  :    ");
		else
			sub_screen_printf(0, 16, "TWOS  :  %2d", now->twos);

		if(now->threes == -1)
			sub_screen_printf(0, 17, "THREES:    ");
		else
			sub_screen_printf(0, 17, "THREES:  %2d", now->threes);

		if(now->fours == -1)
			sub_screen_printf(0, 18, "FOURS :    ");
		else
			sub_screen_printf(0, 18, "FOURS :  %2d", now->fours);

		if(now->fives == -1)
			sub_screen_printf(0, 19, "FIVES :    ");
		else
			sub_screen_printf(0, 19, "FIVES :  %2d", now->fives);

		if(now->sixes == -1)
			sub_screen_printf(0, 20, "SIXES :    ");
		else
			sub_screen_printf(0, 20, "SIXES :  %2d", now->sixes);

		sub_screen_printf(0, 21, "BONUS :  %2s", (TOP_SUBTOTAL >= 63 ? "35" : "  "));
		sub_screen_printf(0, 22, "TOTAL : %3d", (TOP_SUBTOTAL + (TOP_SUBTOTAL >= 63 ? 35 : 0)));

		if(now->three_of_a_kind == -1)
			sub_screen_printf(15, 15, "3 OF A KIND:    ");
		else
			sub_screen_printf(15, 15, "3 OF A KIND:  %2d", now->three_of_a_kind);

		if(now->four_of_a_kind == -1)
			sub_screen_printf(15, 16, "4 OF A KIND:    ");
		else
			sub_screen_printf(15, 16, "4 OF A KIND:  %2d", now->four_of_a_kind);

		if(now->full_house == -1)
			sub_screen_printf(15, 17, "FULL HOUSE :    ");
		else
			sub_screen_printf(15, 17, "FULL HOUSE :  %2d", now->full_house);

		if(now->small_straight == -1)
			sub_screen_printf(15, 18, "SM STRAIGHT:    ");
		else
			sub_screen_printf(15, 18, "SM STRAIGHT:  %2d", now->small_straight);

		if(now->large_straight == -1)
			sub_screen_printf(15, 19, "LG STRAIGHT:    ");
		else
			sub_screen_printf(15, 19, "LG STRAIGHT:  %2d", now->large_straight);

		if(now->yahtzee == -1)
			sub_screen_printf(15, 20, "YAHTZEE    :    ");
		else
			sub_screen_printf(15, 20, "YAHTZEE    : %3d", now->yahtzee);

		if(now->chance == -1)
			sub_screen_printf(15, 21, "CHANCE     :    ");
		else
			sub_screen_printf(15, 21, "CHANCE     :  %2d", now->chance);

		sub_screen_printf(15, 22, "TOTAL      : %3d", BOTTOM_SUBTOTAL);
		sub_screen_printf(15, 23, "GRAND TOTAL: %3d", TOP_SUBTOTAL + (TOP_SUBTOTAL >= 63 ? 35 : 0) + BOTTOM_SUBTOTAL);

		now = now->next;
		console_gotoxy(0, 4);
	}

	// show the current player's score on the main screen
	main_screen_clear();

	main_screen_printf(0, 0, "%s", now->name);

	if(now->ones == -1)
		main_screen_printf(0, 1, "ONES  :    ");
	else
		main_screen_printf(0, 1, "ONES  :  %2d", now->ones);

	if(now->twos == -1)
		main_screen_printf(0, 2, "TWOS  :    ");
	else
		main_screen_printf(0, 2, "TWOS  :  %2d", now->twos);

	if(now->threes == -1)
		main_screen_printf(0, 3, "THREES:    ");
	else
		main_screen_printf(0, 3, "THREES:  %2d", now->threes);

	if(now->fours == -1)
		main_screen_printf(0, 4, "FOURS :    ");
	else
		main_screen_printf(0, 4, "FOURS :  %2d", now->fours);

	if(now->fives == -1)
		main_screen_printf(0, 5, "FIVES :    ");
	else
		main_screen_printf(0, 5, "FIVES :  %2d", now->fives);

	if(now->sixes == -1)
		main_screen_printf(0, 6, "SIXES :    ");
	else
		main_screen_printf(0, 6, "SIXES :  %2d", now->sixes);

	main_screen_printf(0, 7, "BONUS :  %2s", (TOP_SUBTOTAL >= 63 ? "35" : "  "));
	main_screen_printf(0, 8, "TOTAL : %3d", (TOP_SUBTOTAL + (TOP_SUBTOTAL >= 63 ? 35 : 0)));

	if(now->three_of_a_kind == -1)
		main_screen_printf(15, 1, "3 OF A KIND:    ");
	else
		main_screen_printf(15, 1, "3 OF A KIND:  %2d", now->three_of_a_kind);

	if(now->four_of_a_kind == -1)
		main_screen_printf(15, 2, "4 OF A KIND:    ");
	else
		main_screen_printf(15, 2, "4 OF A KIND:  %2d", now->four_of_a_kind);

	if(now->full_house == -1)
		main_screen_printf(15, 3, "FULL HOUSE :    ");
	else
		main_screen_printf(15, 3, "FULL HOUSE :  %2d", now->full_house);

	if(now->small_straight == -1)
		main_screen_printf(15, 4, "SM STRAIGHT:    ");
	else
		main_screen_printf(15, 4, "SM STRAIGHT:  %2d", now->small_straight);

	if(now->large_straight == -1)
		main_screen_printf(15, 5, "LG STRAIGHT:    ");
	else
		main_screen_printf(15, 5, "LG STRAIGHT:  %2d", now->large_straight);

	if(now->yahtzee == -1)
		main_screen_printf(15, 6, "YAHTZEE    :    ");
	else
		main_screen_printf(15, 6, "YAHTZEE    : %3d", now->yahtzee);

	if(now->chance == -1)
		main_screen_printf(15, 7, "CHANCE     :    ");
	else
		main_screen_printf(15, 7, "CHANCE     :  %2d", now->chance);

	main_screen_printf(15, 8, "TOTAL      : %3d", BOTTOM_SUBTOTAL);
	main_screen_printf(15, 9, "GRAND TOTAL: %3d", TOP_SUBTOTAL + (TOP_SUBTOTAL >= 63 ? 35 : 0) + BOTTOM_SUBTOTAL);

	// show the current roll number
	if(roll == 0)
		main_screen_printf(0, 11, "Roll: --");
	else
		main_screen_printf(0, 11, "Roll:  %d", roll);

	// show the actual dice and "roll" button
	for(i = 0; i < 5; i++) {
		if(roll == 0 || mode != MODE_PLAYING) {
			draw_bitmap((6 * i + 1) * 8 + 15, 14 * 8, 9 + dice[i]);
		} else if(held[i]) {
			draw_bitmap((6 * i + 1) * 8 + 15, 14 * 8, 17 + dice[i]);
		} else {
			draw_bitmap((6 * i + 1) * 8 + 15, 14 * 8, 1 + dice[i]);
		}
	}

	if(roll >= 3 || mode != MODE_PLAYING) {
		draw_bitmap(110, 150, 31);
	} else {
		draw_bitmap(110, 150, 30);
	}

	flip();
}

void draw_bitmap(int x, int y, int img)
{
	int colstart, rowstart, width, height;
	int i;

	switch(img)
	{
		case 0: // white upside-down question mark
			colstart = 0;
			rowstart = 0;
			width = 20;
			height = 20;
			break;
		case 1: // white question mark
			colstart = 21;
			rowstart = 0;
			width = 20;
			height = 20;
			break;
		case 2: // white 1 dice
			colstart = 42;
			rowstart = 0;
			width = 20;
			height = 20;
			break;
		case 3: // white 2 dice
			colstart = 63;
			rowstart = 0;
			width = 20;
			height = 20;
			break;
		case 4: // white 3 dice
			colstart = 84;
			rowstart = 0;
			width = 20;
			height = 20;
			break;
		case 5: // white 4 dice
			colstart = 105;
			rowstart = 0;
			width = 20;
			height = 20;
			break;
		case 6: // white 5 dice
			colstart = 126;
			rowstart = 0;
			width = 20;
			height = 20;
			break;
		case 7: // white 6 dice
			colstart = 147;
			rowstart = 0;
			width = 20;
			height = 20;
			break;
		case 8: // gray upside-down question mark
			colstart = 0;
			rowstart = 21;
			width = 20;
			height = 20;
			break;
		case 9: // gray question mark
			colstart = 21;
			rowstart = 21;
			width = 20;
			height = 20;
			break;
		case 10: // gray 1 dice
			colstart = 42;
			rowstart = 21;
			width = 20;
			height = 20;
			break;
		case 11: // gray 2 dice
			colstart = 63;
			rowstart = 21;
			width = 20;
			height = 20;
			break;
		case 12: // gray 3 dice
			colstart = 84;
			rowstart = 21;
			width = 20;
			height = 20;
			break;
		case 13: // gray 4 dice
			colstart = 105;
			rowstart = 21;
			width = 20;
			height = 20;
			break;
		case 14: // gray 5 dice
			colstart = 126;
			rowstart = 21;
			width = 20;
			height = 20;
			break;
		case 15: // gray 6 dice
			colstart = 147;
			rowstart = 21;
			width = 20;
			height = 20;
			break;
		case 16: // blue upside-down question mark
			colstart = 0;
			rowstart = 42;
			width = 20;
			height = 20;
			break;
		case 17: // blue question mark
			colstart = 21;
			rowstart = 42;
			width = 20;
			height = 20;
			break;
		case 18: // blue 1 dice
			colstart = 42;
			rowstart = 42;
			width = 20;
			height = 20;
			break;
		case 19: // blue 2 dice
			colstart = 63;
			rowstart = 42;
			width = 20;
			height = 20;
			break;
		case 20: // blue 3 dice
			colstart = 84;
			rowstart = 42;
			width = 20;
			height = 20;
			break;
		case 21: // blue 4 dice
			colstart = 105;
			rowstart = 42;
			width = 20;
			height = 20;
			break;
		case 22: // blue 5 dice
			colstart = 126;
			rowstart = 42;
			width = 20;
			height = 20;
			break;
		case 23: // blue 6 dice
			colstart = 147;
			rowstart = 42;
			width = 20;
			height = 20;
			break;
		case 24: // host button
			colstart = 168;
			rowstart = 0;
			width = 38;
			height = 13;
			break;
		case 25: // play button
			colstart = 168;
			rowstart = 14;
			width = 38;
			height = 13;
			break;
		case 26: // wifi on button
			colstart = 207;
			rowstart = 0;
			width = 38;
			height = 13;
			break;
		case 27: // wifi off button
			colstart = 207;
			rowstart = 14;
			width = 38;
			height = 13;
			break;
		case 28: // plus button
			colstart = 168;
			rowstart = 28;
			width = 14;
			height = 13;
			break;
		case 29: // minus button
			colstart = 182;
			rowstart = 28;
			width = 14;
			height = 13;
			break;
		case 30: // roll button
			colstart = 207;
			rowstart = 28;
			width = 38;
			height = 13;
			break;
		case 31: // gray roll button
			colstart = 207;
			rowstart = 42;
			width = 38;
			height = 13;
			break;
		case 32: // start button
			colstart = 168;
			rowstart = 42;
			width = 38;
			height = 13;
			break;
		default: // invalid
			return;
	}

	for(i = 0; i < height; i++) {
		dmaCopy((u8*)diceBitmap + (rowstart + i) * 244 + colstart, &bg3map[(y + i) * 256 + x], width);
		DC_InvalidateRange(&bg3map[(y + i) * 256 + x], width);
	}
}

void draw_menu()
{
	LPLOBBY_ROOM rm;
	char *name = NULL;
	int cur = 0;
	int max = 0;
	int i;

	main_screen_clear();

	main_screen_printf(0, 0, "Players on this DS: %2d", num_local_players);
	draw_bitmap(220, 0, 29);
	draw_bitmap(240, 0, 28);

	main_screen_printf(5, 2, "a Local Game");
	draw_bitmap(0, 2 * 8 - 2, 25);
	main_screen_printf(5, 4, "a Wireless Game");
	draw_bitmap(0, 4 * 8 - 2, 24);

	draw_bitmap(210, 3 * 8, (wifi_active ? 26 : 27));

	main_screen_printf(0, 6, "Join a Wireless Game:");

	if(wifi_active) {
		i = 0;
		rm = 0;
		while((rm = LOBBY_GetRoomByGame(rm, YAHTZEE_GAMECODE)) != 0) {
			name = LOBBY_GetRoomName(rm);
			cur = LOBBY_GetUsercountInRoom(rm);
			max = LOBBY_GetMaxUsercountInRoom(rm);

			main_screen_printf(0, 7 + i, "%-24s %2d / %2d", name, cur, max);
			i++;
		}
	}

	flip();
}

void draw_server_screen()
{
	LPLOBBY_USER user;
	int i;

	main_screen_clear();

	for(i = 0; i < LOBBY_GetUsercountInRoom(game_room) && i < MAX_CLIENTS; i++) {
		user = LOBBY_GetRoomUserBySlot(game_room, i);
		main_screen_printf(0, i, "%s (%d)", LOBBY_GetUserName(user), players_per_client[i]);
	}

	draw_bitmap(0, 178, 32);

	flip();
}

void display_final_scores(int *scores)
{
	int i;

	main_screen_clear();

	main_screen_printf(0, 0, "Final Scores:");

	for(i = 0; i < num_players; i++) {
		now = player;
		main_screen_printf(0, i + 1, "%4d %s", *scores, now->name);
		now = now->next;
		scores++;
	}

	flip();

	wait_for_keypress(KEY_A | KEY_B | KEY_X | KEY_Y | KEY_START | KEY_SELECT | KEY_L | KEY_R | KEY_TOUCH);

	end_game();

	mode = MODE_MENU;
}

void handle_input()
{
	touchPosition tp;
	u32 kd;
	int target = -1;

	scanKeys();

	kd = keysDown();

	if(kd & KEY_LID) {
		enter_sleep_mode();
	}

	if(kd & KEY_TOUCH) {
		tp = touchReadXY();
		//iprintf("touch %d,%d\n", tp.px, tp.py);

		if(tp.py < 8 * 8) {
			if(roll > 0 && tp.px < 11 * 8) { // left score column
				if(tp.py >= 1 * 8 && tp.py < 2 * 8 && now->ones == -1) {
					target = SCORE_ONES;
				} else if(tp.py >= 2 * 8 && tp.py < 3 * 8 && now->twos == -1) {
					target = SCORE_TWOS;
				} else if(tp.py >= 3 * 8 && tp.py < 4 * 8 && now->threes == -1) {
					target = SCORE_THREES;
				} else if(tp.py >= 4 * 8 && tp.py < 5 * 8 && now->fours == -1) {
					target = SCORE_FOURS;
				} else if(tp.py >= 5 * 8 && tp.py < 6 * 8 && now->fives == -1) {
					target = SCORE_FIVES;
				} else if(tp.py >= 6 * 8 && tp.py < 7 * 8 && now->sixes == -1) {
					target = SCORE_SIXES;
				}
			} else if(roll > 0 && tp.px >= 15 * 8) { // right score column
				if(tp.py >= 1 * 8 && tp.py < 2 * 8 && now->three_of_a_kind == -1) {
					target = SCORE_THREE_OF_A_KIND;
				} else if(tp.py >= 2 * 8 && tp.py < 3 * 8 && now->four_of_a_kind == -1) {
					target = SCORE_FOUR_OF_A_KIND;
				} else if(tp.py >= 3 * 8 && tp.py < 4 * 8 && now->full_house == -1) {
					target = SCORE_FULL_HOUSE;
				} else if(tp.py >= 4 * 8 && tp.py < 5 * 8 && now->small_straight == -1) {
					target = SCORE_SMALL_STRAIGHT;
				} else if(tp.py >= 5 * 8 && tp.py < 6 * 8 && now->large_straight == -1) {
					target = SCORE_LARGE_STRAIGHT;
				} else if(tp.py >= 6 * 8 && tp.py < 7 * 8 && now->yahtzee == -1) {
					target = SCORE_YAHTZEE;
				} else if(tp.py >= 7 * 8 && tp.py < 8 * 8 && now->chance == -1) {
					target = SCORE_CHANCE;
				}
			}

			if(target != -1) {
				if(i_am_server) {
					score_turn(target);

					if(num_clients > 1) {
						memset(send_buffer, 0, MAX_DATA);
						send_buffer[0] = MESSAGE_SCORE_TURN;
						send_buffer[1] = target;
						LOBBY_SendToRoom(game_room, 0x0001, send_buffer, MAX_DATA);
					}

					advance_player();
				} else {
					score_turn(target);

					memset(send_buffer, 0, MAX_DATA);
					send_buffer[0] = MESSAGE_SCORE_TURN;
					send_buffer[1] = target;
					LOBBY_SendToRoom(game_room, 0x0001, send_buffer, MAX_DATA);
				}
			}
		} else if(tp.py >= 14 * 8 && tp.py < 17 * 8) { // dice
			if(roll > 0 && roll < 3) {
				if(tp.px >= 1 * 8 + 15 && tp.px < 1 * 8 + 35)
					held[0] = !held[0];
				else if(tp.px >= 7 * 8 + 15 && tp.px < 7 * 8 + 35)
					held[1] = !held[1];
				else if(tp.px >= 13 * 8 + 15 && tp.px < 13 * 8 + 35)
					held[2] = !held[2];
				else if(tp.px >= 19 * 8 + 15 && tp.px < 19 * 8 + 35)
					held[3] = !held[3];
				else if(tp.px >= 25 * 8 + 15 && tp.px < 25 * 8 + 35)
					held[4] = !held[4];
			}
		} else if(tp.px >= 110 && tp.px < 150 && tp.py >= 150 && tp.py < 165) { // roll
			if(roll < 3) {
				roll_dice();
			}
		}
	}
}

void handle_menu_input()
{
	touchPosition tp;
	u32 kd;
	LPLOBBY_ROOM rm;
	int i;

	scanKeys();

	kd = keysDown();

	if(kd & KEY_LID) {
		enter_sleep_mode();
	}

	if(kd & KEY_TOUCH) {
		tp = touchReadXY();
		//iprintf("touch %d,%d\n", tp.px, tp.py);

		if(tp.py < 13 && tp.px >= 220 && tp.px < 234) { // minus
			if(num_local_players > 0)
				num_local_players--;
		} else if(tp.py < 13 && tp.px >= 240 && tp.px < 254) { // plus
			if(num_local_players < 99)
				num_local_players++;
		} else if(tp.py >= 2 * 8 - 2 && tp.py < 3 * 8 && tp.px < 40) { // local game
			start_game(1);
		} else if(tp.py >= 4 * 8 - 2 && tp.py < 5 * 8 && tp.px < 40) { // host game
			if(wifi_active) {
				host_game();
				mode = MODE_NEED_CLIENTS;
			} else {
				iprintf("You must turn on WiFi in order\nto host or join a Wireles game!\n");
			}
		} else if(tp.px >= 210 && tp.py >= 3 * 8 && tp.py <= 3 * 8 + 13) { // wifi on/off
			wifi_active = !wifi_active;
			if(wifi_active && !lobby_ever_init)
				activate_wifi();
		} else if(tp.py >= 7 * 8 && wifi_active) { // join game
			i = 0;
			rm = 0;
			while((rm = LOBBY_GetRoomByGame(rm, YAHTZEE_GAMECODE)) != 0) {
				if(i == tp.py / 8 - 7) {
					join_game(rm);
					mode = MODE_WAITING;
					break;
				}
				i++;
			}
		}
	}
}

void handle_server_input()
{
	touchPosition tp;
	u32 kd;

	scanKeys();

	kd = keysDown();

	if(kd & KEY_LID) {
		enter_sleep_mode();
	}

	if(kd & KEY_TOUCH) {
		tp = touchReadXY();
		//iprintf("touch %d,%d\n", tp.px, tp.py);

		if(tp.py >= 178 && tp.px < 40) {
			start_game(0);
		}
	}
}

int animate_dice(time_t start)
{
	int i, diff;
	static int toggle = 0;

	// we're not rolling; nothing to do
	if(start == 0)
		return 0;

	diff = time(0) - start;

	// we've rolled long enough; done animating
	if(diff >= 2 || toggle >= 100) {
		toggle = 0;
		return 0;
	}

	// still rolling
	for(i = 0; i < 5; i++) {
		if(!held[i]) {
			if(mode != MODE_PLAYING) {
				draw_bitmap((6 * i + 1) * 8 + 15, 14 * 8, 8 + (toggle / 10) % 2);
			} else {
				draw_bitmap((6 * i + 1) * 8 + 15, 14 * 8, 0 + (toggle / 10) % 2);
			}
		}
	}

	flip();

	toggle++;

	return 1;
}

void activate_wifi()
{
	// Lobby stuff
	IPC_Init();
	IPC_SetChannelCallback(0, &LWIFI_IPC_Callback) ;
	//	IPC_SetMsgCompleteCallback(&IPC_RcvCompleteCheck);
	LOBBY_Init() ;
	LOBBY_SetStreamHandler(0x0001, &receiveUserData);
	LOBBY_SetUserInfoCallback(&userCallback);
	lobby_ever_init = 1;
	wifi_active = 1;
}

//---------------------------------------------------------------------------------
int main(void)
{
	powerON(POWER_ALL);
	allowed_to_sleep = 1;

	// configure graphics
	vramSetMainBanks(VRAM_A_MAIN_BG, VRAM_B_LCD, VRAM_C_SUB_BG, VRAM_D_LCD);
	videoSetMode(MODE_5_2D | DISPLAY_BG0_ACTIVE | DISPLAY_BG3_ACTIVE);
	videoSetModeSub(MODE_0_2D | DISPLAY_BG0_ACTIVE);
	BG0_CR = BG_32x32 | BG_COLOR_16 | BG_MAP_BASE(0) | BG_TILE_BASE(1);
	BG3_CR = BG_BMP8_256x256 | BG_BMP_BASE(2);
	BG3_XDX = 1 << 8;
	BG3_XDY = 0;
	BG3_YDX = 0;
	BG3_YDY = 1 << 8;
	BG3_CX = 0;
	BG3_CY = 0;

	// configure palette
	dmaCopy(dicePal, (u16*)BG_PALETTE, dicePalLen);
	BG_PALETTE[0] = RGB15(0,0,0);
	BG_PALETTE[15] = RGB15(20,20,31);
	BG_PALETTE[31] = RGB15(15,15,15);
	BG_PALETTE[255] = RGB15(31,31,31);

	// standard console on sub screen
	SUB_BG0_CR = BG_MAP_BASE(31);
	BG_PALETTE_SUB[255] = RGB15(31,31,31);
	consoleInitDefault((u16*)SCREEN_BASE_BLOCK_SUB(31), (u16*)CHAR_BASE_BLOCK_SUB(0), 16);

	// copy console font to main screen as well
	DC_FlushRange((u8*)CHAR_BASE_BLOCK_SUB(0), 256 * 16 * 2);
	dmaCopy((u8*)CHAR_BASE_BLOCK_SUB(0), (u8*)CHAR_BASE_BLOCK(1), 256 * 16 * 2);

	// enable multiple (16) color palettes for the console font
	u32 *v;
	for(v = (u32*)CHAR_BASE_BLOCK_SUB(0); v < (u32*)CHAR_BASE_BLOCK_SUB(1); v++)
		*v |= 0xEEEEEEEE;

	defaultExceptionHandler();

	// VBlank IRQ handler
	irqInit();
	irqSet(IRQ_VBLANK, &VBlankIRQ);
	irqEnable(IRQ_VBLANK);

	srand(time(NULL));

	lcdMainOnBottom();

	main_screen_clear();
	flip();

	// welcome
	iprintf("YahtzeeDS v" VERSION "\n");
	iprintf("by Dopefish\n");
	iprintf("http://vespenegas.com\n\n");

	num_local_players = 1;

	while(1) {
		if(!animate_dice(dice_animation_start)) {
			dice_animation_start = 0;

			switch(mode) {
				case MODE_MENU:
					draw_menu();
					handle_menu_input();
					break;
				case MODE_PLAYING:
					draw_scoresheet();
					handle_input();
					break;
				case MODE_WAITING:
					if(now != NULL)
						draw_scoresheet();
					break;
				case MODE_NEED_CLIENTS:
					draw_server_screen();
					handle_server_input();
					break;
			}
		}

		// don't sleep if we're in the middle of a multi-player game, because it'll hose everyone else
		if(mode != MODE_MENU && num_clients > 0) {
			allowed_to_sleep = 0;
		} else {
			allowed_to_sleep = 1;
		}

		swiWaitForVBlank();
	}

	return 0;
}
